import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from torch.autograd import Variable


class ULSTM2(nn.Module):
    def __init__(self, input_dim, hidden_dim, num_layers, fc_dims, output_dim):
        super(ULSTM2, self).__init__()
        self.lstm1 = nn.LSTM(input_dim, hidden_dim, 1, batch_first=True)
        self.bilstm2 = nn.LSTM(768, hidden_dim, num_layers, batch_first=True, bidirectional=True)
        self.bilstm = nn.LSTM(input_dim, hidden_dim, num_layers, batch_first=True, bidirectional=True)
        self.fc_layers = []
        prev_dim = 2 * hidden_dim
        for fc_dim in fc_dims:
            self.fc_layers.append(nn.Linear(prev_dim, fc_dim))
            self.fc_layers.append(nn.Dropout(0.8))
            self.fc_layers.append(nn.ReLU(True))
            prev_dim = fc_dim
        self.fc_layer_module = nn.ModuleList(self.fc_layers)
        self.last_fc = nn.Linear(prev_dim, output_dim)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x, x_lengths):
        packed_sequence = torch.nn.utils.rnn.pack_padded_sequence(x, x_lengths, batch_first=True)
        lstm_output1, _ = self.lstm1(packed_sequence)
        lstm_output2, _ = self.bilstm(packed_sequence)
        out1, _ = torch.nn.utils.rnn.pad_packed_sequence(lstm_output1, batch_first=True)
        out2, _ = torch.nn.utils.rnn.pad_packed_sequence(lstm_output2, batch_first=True)
        out = torch.cat((out1, out2), 2)
        packed_sequence = torch.nn.utils.rnn.pack_padded_sequence(out, x_lengths, batch_first=True)
        bilstm_output, _ = self.bilstm2(packed_sequence)
        out, _ = torch.nn.utils.rnn.pad_packed_sequence(bilstm_output, batch_first=True)
        for layer in self.fc_layer_module:
            out = layer(out)
        out = self.last_fc(out)
        out = self.sigmoid(out)
        return out
