import json
import random
import numpy as np

from utils import read_json, AugmentedList
from utils import PHYS_DICT, AA_DICT, BLOSUM_DICT, SS_DICT

from torch.autograd import Variable
from torch import FloatTensor, LongTensor

TRAIN_MODE = 0
DEV_MODE = 1
TEST_MODE = 2


# Dataset Class
class Dataset:
    def __init__(self, input_types=['ONEHOT'], output_types=['PHI', 'PSI'], readfile=[]):
        self.input_types = input_types
        self.output_types = output_types

        # Calculate input_dim
        input_dim = 0
        for input_type in input_types:
            if input_type == 'ONEHOT': input_dim += 20
            if input_type == 'BLOSUM': input_dim += 20
            if input_type == 'PHYS': input_dim += 7
            if input_type == 'PSSM': input_dim += 20
            if input_type == 'PC': input_dim += 20
            if input_type == 'PSFM': input_dim += 20
            if input_type == 'EC': input_dim += 400
            if input_type == 'RE': input_dim += 20
            if input_type == 'DC': input_dim += 1
            if input_type == 'PHIPSI': input_dim += 4
            if input_type == 'RAMA': input_dim += 144

        self.input_dim = input_dim

        # Calculate output_dim
        self.output_dim = 2 * len(output_types)
        dirt_json = readfile
        dirt = 'features/'
        train_data = read_json(dirt_json + 'train.json')
        dev_data = read_json(dirt_json + 'dev.json')
        test_data = read_json(dirt_json + 'test.json')

        self.train = AugmentedList(list(train_data.items()), shuffle_between_epoch=True)
        self.dev = AugmentedList(list(dev_data.items()))
        self.test = AugmentedList(list(test_data.items()))

        # Read features file
        self.pssm, self.pc, self.ec, self.psfm, self.re, self.dc, self.phipsi, self.rama = {}, {}, {}, {}, {}, {}, {}, {}

        try:
            if 'PSSM' in input_types:
                for i in range(6):
                    pssm_i =read_json(dirt_json + dirt + 'pssm_' + str(i + 1) + '.json')
                    for k, v in pssm_i.items():
                        self.pssm[k] = v
            if 'PSFM' in input_types:
                for i in range(6):
                    psfm_i = read_json(dirt_json + dirt + 'psfm_' + str(i + 1) + '.json')
                    for k, v in psfm_i.items():
                        self.psfm[k] = v
            if 'PC' in input_types:
                for i in range(6):
                    pc_i = read_json(dirt_json + dirt + 'pc_' + str(i + 1) + '.json')
                    for k, v in pc_i.items():
                        self.pc[k] = v
            if 'RE' in input_types:
                for i in range(6):
                    re_i = read_json(dirt_json + dirt + 're_' + str(i + 1) + '.json')
                    for k, v in re_i.items():
                        self.re[k] = v
            if 'DC' in input_types:
                for i in range(6):
                    dc_i = read_json(dirt_json + dirt + 'dc_' + str(i + 1) + '.json')
                    for k, v in dc_i.items():
                        self.dc[k] = v
            if 'EC' in input_types:
                for i in range(6):
                    ec_i = read_json(dirt_json + dirt + 'ec_' + str(i + 1) + '.json')
                    for k, v in ec_i.items():
                        self.ec[k] = v
            if 'PHIPSI' in input_types:
                for i in range(6):
                    phipsi_i = read_json(dirt_json + dirt + 'phipsi_' + str(i + 1) + '.json')
                    for k, v in phipsi_i.items():
                        self.phipsi[k] = v
            if 'RAMA' in input_types:
                for i in range(6):
                    rama_i = read_json(dirt_json + dirt + 'rama_' + str(i + 1) + '.json')
                    for k, v in rama_i.items():
                        self.rama[k] = v
        except:
            if 'PSSM' in input_types:
                    pssm_i =read_json(dirt_json + dirt + 'pssm_1.json')
                    for k, v in pssm_i.items():
                        self.pssm[k] = v
            if 'PSFM' in input_types:
                    psfm_i = read_json(dirt_json + dirt + 'psfm_1.json')
                    for k, v in psfm_i.items():
                        self.psfm[k] = v
            if 'PC' in input_types:
                    pc_i = read_json(dirt_json + dirt + 'pc_1.json')
                    for k, v in pc_i.items():
                        self.pc[k] = v
            if 'RE' in input_types:
                    re_i = read_json(dirt_json + dirt + 're_1.json')
                    for k, v in re_i.items():
                        self.re[k] = v
            if 'DC' in input_types:
                    dc_i = read_json(dirt_json + dirt + 'dc_1.json')
                    for k, v in dc_i.items():
                        self.dc[k] = v
            if 'EC' in input_types:
                    ec_i = read_json(dirt_json + dirt + 'ec_1.json')
                    for k, v in ec_i.items():
                        self.ec[k] = v
            if 'PHIPSI' in input_types:
                    phipsi_i = read_json(dirt_json + dirt + 'phipsi_1.json')
                    for k, v in phipsi_i.items():
                        self.phipsi[k] = v
            if 'RAMA' in input_types:
                    rama_i = read_json(dirt_json + dirt + 'rama_1.json')
                    for k, v in rama_i.items():
                        self.rama[k] = v

    def next_batch(self, batch_size, mode=TRAIN_MODE):
        if mode == TRAIN_MODE: examples = self.train.next_items(batch_size)
        elif mode == DEV_MODE: examples = self.dev.next_items(batch_size)
        elif mode == TEST_MODE: examples = self.test.next_items(batch_size)

        lengths = [len(example[1]['seq']) for example in examples]
        max_len = max(lengths) #167
        # Create output masks
        masks = []
        for length in lengths:
            mask = [[1] * self.output_dim] * length + [[0] * self.output_dim] * (max_len - length)
            masks.append(mask)
        masks = Variable(FloatTensor(masks))

        row = 0
        ids, inputs, outputs, angles = [], [], [], []
        for example in examples:
            _id = example[0]
            ids.append(_id)
            seq = example[1]['seq']
            seq_length = len(seq)
            phi, psi, ss = example[1]['phi'], example[1]['psi'], example[1]['ss']
            pssms = self.pssm.get(_id, [[0] * 20] * len(seq))
            pcs = self.pc.get(_id, [[0] * 20] * len(seq))
            psfms = self.psfm.get(_id, [[0] * 20] * len(seq))
            res = self.re.get(_id, [[0] * 20] * len(seq))
            dcs = self.dc.get(_id, [[0] * 1] * len(seq))
            ecs = self.ec.get(_id, [[0] * 400] * len(seq))
            phipsis = self.phipsi.get(_id, [[0] * 4] * len(seq))
            ramas = self.rama.get(_id, [[0] * 144] * len(seq))

            # Construct _inputs
            _inputs = []
            for i in range(seq_length):
                _input = []
                for input_type in self.input_types:
                    if input_type == 'ONEHOT': _input += AA_DICT[seq[i]]
                    if input_type == 'BLOSUM': _input += BLOSUM_DICT[seq[i]]
                    if input_type == 'PHYS': _input += PHYS_DICT[seq[i]]
                    if input_type == 'PSSM': _input += pssms[i]
                    if input_type == 'PC': _input += pcs[i]
                    if input_type == 'RE': _input += res[i]
                    if input_type == 'DC': _input += dcs[i]
                    if input_type == 'PSFM': _input += psfms[i]
                    if input_type == 'EC': _input += ecs[i]
                    if input_type == 'PHIPSI': _input += phipsis[i]
                    if input_type == 'RAMA': _input += ramas[i]
                _inputs.append(_input)
            _inputs += [[0] * self.input_dim] * (max_len - len(_inputs))
            inputs.append(_inputs)

            # Construct _angles and _outputs
            _angles, _outputs = [], []
            for i in range(seq_length):
                _angle, _output = [], []
                j = 0
                for output_type in self.output_types:
                    if output_type == 'PHI': angle = np.float(phi[i])
                    if output_type == 'PSI': angle = np.float(psi[i])
                    _angle.append(angle)
                    _output += [np.sin(np.deg2rad(angle)), np.cos(np.deg2rad(angle))]
                    if angle == 360.0:
                        masks[row, i, j * 2] = 0
                        masks[row, i, j * 2 + 1] = 0
                    j += 1
                _outputs.append(_output)
                _angles.append(_angle)
            _outputs += [[0] * self.output_dim] * (max_len - len(_outputs))
            angles.append(_angles)
            outputs.append(_outputs)

            row += 1

        # Sorted in decreasing order of length
        for i in range(batch_size):
            for j in range(i+1, batch_size):
                if lengths[i] < lengths[j]:
                    ids[i], ids[j] = ids[j], ids[i]
                    lengths[i], lengths[j] = lengths[j], lengths[i]
                    inputs[i], inputs[j] = inputs[j], inputs[i]
                    outputs[i], outputs[j] = outputs[j], outputs[i]
                    x_row = masks[i,:, :].clone()
                    masks[i,:, :] = masks[j,:, :]
                    masks[j,:, :] = x_row

        # Convert to FloatTensors
        inputs = Variable(FloatTensor(inputs))
        outputs = Variable(FloatTensor(outputs))

        # scale outputs to 0 to 1 range. sigmoid activation.
        outputs = (outputs + 1.0) / 2.0

        return ids, lengths, inputs, outputs, masks, angles
